import type { IBuilderObj, IModelEntity, IPaginated, IQueryFields, QueryBuilder, PrismaValidStringFilter, PrismaWhereInput } from '../types';
declare class ApiQueryBuilder<TModelEntity extends IModelEntity = IModelEntity, WhereInput = PrismaWhereInput, PrismaModelDelegate extends any = any> {
    protected query: IQueryFields;
    protected modelEntity: PrismaModelDelegate;
    protected entity: IModelEntity;
    protected validStringFilter: PrismaValidStringFilter[];
    protected builder: QueryBuilder;
    protected builderObj: IBuilderObj;
    /**
     * The constructor takes in a modelEntity, entity, query, and validStringFilter
     * @param {PrismaModelDelegate} modelEntity - This is the prisma entity e.g this.prisma.shoppingItem
     * @param {TModelEntity} entity - This is the entity that you want to query. it also contains all the settings for filtering and selecting of data
     * @param {IQueryFields} [query] - This is the query string fields from the api.
     * @param {PrismaValidStringFilter[]} validStringFilter - This is an array of
     * PrismaValidStringFilter. This is used to validate the string filter.
     */
    constructor(modelEntity: PrismaModelDelegate, entity: TModelEntity, query?: IQueryFields, validStringFilter?: PrismaValidStringFilter[]);
    /**
     * It takes a query string, parses it into an object, and then recursively filters out any invalid
     * keys
     * @returns The builder object with the filter property set to the validFilterObj.
     */
    filter(): this;
    /**
     * It takes the sort query parameter, splits it into an array, and then filters it to only include
     * the sortable columns
     * @returns The builder object
     */
    sort(): this;
    /**
     * It takes the `select` query parameter, parses it into an array, and then converts that array into
     * an object with the keys being the column names and the values being `true` or `false` depending on
     * whether the column was selected or not
     * @returns The builder object
     */
    select(): this;
    /**
     * If the user has passed in a query parameter called `include` then we will add the `select`
     * property to the knex query builder
     * @returns The builder object with the select property set to the selectables object.
     */
    populate(): this;
    /**
     * It takes the page and limit from the query string, and then uses them to calculate the skip and
     * take values for the pagination
     * @returns The builder object with the paginate property set to an object with skip and take
     * properties.
     */
    paginate(): this;
    /**
     * It takes the filter, paginate, select, and sort properties from the builder object and assigns
     * them to the builderObj property
     * @returns The builder object
     */
    build(): this;
    /**
     * This function sorts, filters, selects, populates, and paginates the query, and then returns the
     * query.
     * @returns The query object.
     */
    all(): this;
    /**
     * It returns a paginated object with the data and the paginated object
     * @returns An object with two properties: data and paginated.
     */
    many(): Promise<"NotFoundException" | {
        data: Partial<TModelEntity>[];
        paginated: IPaginated;
    }>;
    /**
     * It takes a filter object as a parameter, checks if the filter object has a where property, if it
     * doesn't, it returns an error message, if it does, it queries the database using the filter object,
     * counts the total number of records that match the filter, calculates the total number of pages,
     * and returns the data and paginated object
     * @param {WhereInput} filter - WhereInput - This is the filter object that is passed to the findMany
     * method.
     * @returns - An object with two properties:
     *     - data: An array of objects that match the filter criteria
     *     - paginated: An object with the following properties:
     *       - currentPage: The current page number
     *       - perPage: The number of records per page
     *       - totalPages: The total number of pages
     *       - totalRecords: The total number
     */
    findMany(filter: WhereInput): Promise<"NotFoundException" | "Invalid Filter Parameter" | {
        data: Partial<TModelEntity>[];
        paginated: IPaginated;
    }>;
    /**
     * It takes a filter object as a parameter, checks if the filter object has a where property, if it
     * doesn't, it returns an error message, if it does, it creates a new object with the where property
     * of the filter object and the where property of the builder object, and then it returns the result
     * of the findFirst method of the modelEntity object
     * @param {WhereInput} filter - WhereInput
     * @returns The first record that matches the filter.
     */
    one(filter: WhereInput): Promise<Partial<TModelEntity> | "Invalid Filter Parameter">;
    /**
     * If the value is undefined or null, return an empty array. If the value is an array, return the
     * array with any undefined values removed. If the value is not an array, return an array with the
     * value as the only element
     * @param {any | any[]} val - any | any[]
     * @returns An array of values.
     */
    private toArray;
    /**
     * It takes an array of strings and returns an object with the same strings as keys and boolean
     * values
     * @param {string[]} selectables - string[] - an array of strings that are the selectable options
     * @param {string[]} array - string[] - the array of strings to convert to a boolean object
     * @returns An object with the keys of the selectables array and the values of the array.
     */
    private arrayToBoolObject;
    /**
     * It takes a filter object, picks only the valid keys, and then recursively filters the object until
     * it's empty
     * @param {object} filterObj - object - The object that you want to filter.
     */
    private filterRecursive;
}
export default ApiQueryBuilder;
