"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const qs = require("qs");
const query_string_parser_1 = require("./query-string-parser");
const query_builder_constant_1 = require("./query-builder.constant");
/* It takes a model entity, a query object and a valid string filter array and returns a valid Prisma
query object */
class ApiQueryBuilder {
    /**
     * The constructor takes in a modelEntity, entity, query, and validStringFilter
     * @param {PrismaModelDelegate} modelEntity - This is the prisma entity e.g this.prisma.shoppingItem
     * @param {TModelEntity} entity - This is the entity that you want to query. it also contains all the settings for filtering and selecting of data
     * @param {IQueryFields} [query] - This is the query string fields from the api.
     * @param {PrismaValidStringFilter[]} validStringFilter - This is an array of
     * PrismaValidStringFilter. This is used to validate the string filter.
     */
    constructor(modelEntity, entity, query, validStringFilter = query_builder_constant_1.defaultPrismaStringFilter) {
        /* Defining the default values for the query builder. */
        this.builder = {
            filter: {},
            paginated: { currentPage: 1, pages: 1 },
            select: {},
            sort: {},
            paginate: { skip: query_builder_constant_1.skip, take: query_builder_constant_1.take }
        };
        /* Defining the default values for the query builder object. */
        this.builderObj = {
            skip: query_builder_constant_1.skip,
            take: query_builder_constant_1.take,
            where: {},
            select: {},
            orderBy: {}
        };
        this.modelEntity = modelEntity;
        this.query = query !== null && query !== void 0 ? query : query_builder_constant_1.defaultQueryFields;
        this.entity = entity;
        this.validStringFilter = validStringFilter;
    }
    /**
     * It takes a query string, parses it into an object, and then recursively filters out any invalid
     * keys
     * @returns The builder object with the filter property set to the validFilterObj.
     */
    filter() {
        var _a, _b;
        const filterObj = (0, query_string_parser_1.parseObject)(qs.parse((_b = (_a = this === null || this === void 0 ? void 0 : this.query) === null || _a === void 0 ? void 0 : _a.filter) !== null && _b !== void 0 ? _b : '', { comma: true }));
        const validFilterObj = this.filterRecursive(filterObj);
        this.builder.filter = validFilterObj;
        return this;
    }
    /**
     * It takes the sort query parameter, splits it into an array, and then filters it to only include
     * the sortable columns
     * @returns The builder object
     */
    sort() {
        var _a, _b;
        const sort = _.intersectionWith(((_b = (_a = this === null || this === void 0 ? void 0 : this.query) === null || _a === void 0 ? void 0 : _a.sort) !== null && _b !== void 0 ? _b : query_builder_constant_1.defaultQueryFields.sort).split(','), [
            ...this.entity.sortableColumns,
            ...this.entity.sortableColumns.map((col) => `-${col}`),
        ]);
        this.builder.sort = sort === null || sort === void 0 ? void 0 : sort.map((s) => {
            const key = (s === null || s === void 0 ? void 0 : s.startsWith('-')) ? s === null || s === void 0 ? void 0 : s.slice(1, s === null || s === void 0 ? void 0 : s.length) : s;
            return { [key]: (s === null || s === void 0 ? void 0 : s.startsWith('-')) ? 'desc' : 'asc' };
        });
        return this;
    }
    /**
     * It takes the `select` query parameter, parses it into an array, and then converts that array into
     * an object with the keys being the column names and the values being `true` or `false` depending on
     * whether the column was selected or not
     * @returns The builder object
     */
    select() {
        var _a;
        const selectObj = qs.parse(`query=${((_a = this === null || this === void 0 ? void 0 : this.query) === null || _a === void 0 ? void 0 : _a.select) || 'true'}`, {
            comma: true,
        });
        this.builder.select = this.arrayToBoolObject(this.entity.selectableColumns, this.toArray(selectObj === null || selectObj === void 0 ? void 0 : selectObj.query));
        return this;
    }
    /**
     * If the user has passed in a query parameter called `include` then we will add the `select`
     * property to the knex query builder
     * @returns The builder object with the select property set to the selectables object.
     */
    populate() {
        var _a, _b, _c;
        const includeObj = qs.parse((_b = (_a = this === null || this === void 0 ? void 0 : this.query) === null || _a === void 0 ? void 0 : _a.include) !== null && _b !== void 0 ? _b : '', { comma: true });
        if ((0, query_builder_constant_1.isObjectEmpty)(includeObj))
            return this;
        /* Checking if the user has passed in a query parameter called `include` then we will add the
         * `select`
        ** property to the knex query builder */
        (_c = this.entity.allowedRelationShips) === null || _c === void 0 ? void 0 : _c.forEach((col) => {
            var _a;
            const selectables = this.entity.selectableRelationShips[col];
            const includeQueryParam = this.toArray((_a = includeObj[col]) !== null && _a !== void 0 ? _a : []);
            /* Filtering the include query parameter to only include the selectable columns. */
            const filteredIncludeObj = this.arrayToBoolObject(selectables, includeQueryParam);
            if (!this.builder.select)
                this.builder.select = {};
            // if true select only the selectables else select the valid relational columns
            const select = includeQueryParam[0] == 'true' ?
                this.arrayToBoolObject(selectables, selectables)
                : (0, query_builder_constant_1.isObjectEmpty)(filteredIncludeObj) ? undefined : filteredIncludeObj;
            this.builder.select[col] = { select };
        });
        return this;
    }
    /**
     * It takes the page and limit from the query string, and then uses them to calculate the skip and
     * take values for the pagination
     * @returns The builder object with the paginate property set to an object with skip and take
     * properties.
     */
    paginate() {
        var _a, _b, _c, _d;
        const page = ((_b = (_a = this.query) === null || _a === void 0 ? void 0 : _a.page) !== null && _b !== void 0 ? _b : query_builder_constant_1.defaultOptions.currentPage) * 1 || 1;
        const limit = ((_d = (_c = this.query) === null || _c === void 0 ? void 0 : _c.limit) !== null && _d !== void 0 ? _d : query_builder_constant_1.defaultOptions.perPage) * 1 || 50;
        const skip = (page - 1) * limit;
        this.builder.paginate = { skip, take: limit };
        return this;
    }
    /**
     * It takes the filter, paginate, select, and sort properties from the builder object and assigns
     * them to the builderObj property
     * @returns The builder object
     */
    build() {
        const { filter, paginate, select, sort } = this.builder;
        this.builderObj = Object.assign({ where: filter, select, orderBy: sort }, paginate);
        return this;
    }
    /**
     * This function sorts, filters, selects, populates, and paginates the query, and then returns the
     * query.
     * @returns The query object.
     */
    all() {
        this.sort();
        this.filter();
        this.select();
        this.populate();
        this.paginate();
        return this;
    }
    /**
     * It returns a paginated object with the data and the paginated object
     * @returns An object with two properties: data and paginated.
     */
    async many() {
        var _a, _b, _c;
        const data = await this.modelEntity.findMany(this.builderObj);
        const totalRecords = await this.modelEntity.count({
            where: this.builderObj.where,
        });
        const paginated = {
            currentPage: (_b = (_a = this.query) === null || _a === void 0 ? void 0 : _a.page) !== null && _b !== void 0 ? _b : query_builder_constant_1.defaultQueryFields.page * 1,
            perPage: ((_c = this.query.limit) !== null && _c !== void 0 ? _c : query_builder_constant_1.defaultQueryFields.limit) * 1 || (data === null || data === void 0 ? void 0 : data.length),
            totalPages: Math.ceil(totalRecords / this.builderObj.take),
            totalRecords,
        };
        if (paginated.currentPage > 1 && paginated.currentPage > paginated.totalPages)
            return 'NotFoundException';
        return { data, paginated };
    }
    /**
     * It takes a filter object as a parameter, checks if the filter object has a where property, if it
     * doesn't, it returns an error message, if it does, it queries the database using the filter object,
     * counts the total number of records that match the filter, calculates the total number of pages,
     * and returns the data and paginated object
     * @param {WhereInput} filter - WhereInput - This is the filter object that is passed to the findMany
     * method.
     * @returns - An object with two properties:
     *     - data: An array of objects that match the filter criteria
     *     - paginated: An object with the following properties:
     *       - currentPage: The current page number
     *       - perPage: The number of records per page
     *       - totalPages: The total number of pages
     *       - totalRecords: The total number
     */
    async findMany(filter) {
        var _a, _b, _c;
        if (!(filter === null || filter === void 0 ? void 0 : filter.where))
            return 'Invalid Filter Parameter';
        const data = await this.modelEntity.findMany(filter);
        const totalRecords = await this.modelEntity.count({
            where: filter === null || filter === void 0 ? void 0 : filter.where,
        });
        const paginated = {
            currentPage: (_b = (_a = this.query) === null || _a === void 0 ? void 0 : _a.page) !== null && _b !== void 0 ? _b : query_builder_constant_1.defaultQueryFields.page * 1,
            perPage: ((_c = this.query.limit) !== null && _c !== void 0 ? _c : query_builder_constant_1.defaultQueryFields.limit) * 1 || (data === null || data === void 0 ? void 0 : data.length),
            totalPages: Math.ceil(totalRecords / this.builderObj.take),
            totalRecords,
        };
        if (paginated.currentPage > 1 && paginated.currentPage > paginated.totalPages)
            return 'NotFoundException';
        return { data, paginated };
    }
    /**
     * It takes a filter object as a parameter, checks if the filter object has a where property, if it
     * doesn't, it returns an error message, if it does, it creates a new object with the where property
     * of the filter object and the where property of the builder object, and then it returns the result
     * of the findFirst method of the modelEntity object
     * @param {WhereInput} filter - WhereInput
     * @returns The first record that matches the filter.
     */
    async one(filter) {
        if (!(filter === null || filter === void 0 ? void 0 : filter.where))
            return 'Invalid Filter Parameter';
        const _a = this.builderObj, { where, skip, take } = _a, obj = __rest(_a, ["where", "skip", "take"]);
        return this.modelEntity.findFirst(Object.assign({ where: Object.assign(Object.assign({}, filter === null || filter === void 0 ? void 0 : filter.where), where) }, obj));
    }
    /**
     * If the value is undefined or null, return an empty array. If the value is an array, return the
     * array with any undefined values removed. If the value is not an array, return an array with the
     * value as the only element
     * @param {any | any[]} val - any | any[]
     * @returns An array of values.
     */
    toArray(val) {
        if (val == undefined || val == null)
            return [];
        if (Array.isArray(val))
            return val.filter((col) => col != undefined);
        return [val];
    }
    /**
     * It takes an array of strings and returns an object with the same strings as keys and boolean
     * values
     * @param {string[]} selectables - string[] - an array of strings that are the selectable options
     * @param {string[]} array - string[] - the array of strings to convert to a boolean object
     * @returns An object with the keys of the selectables array and the values of the array.
     */
    arrayToBoolObject(selectables, array) {
        if (array[0] == 'true')
            return this.arrayToBoolObject(selectables, selectables);
        array = _.intersectionWith(array, selectables);
        const obj = {};
        array === null || array === void 0 ? void 0 : array.forEach((element) => { obj[element] = true; });
        return (0, query_builder_constant_1.isObjectEmpty)(obj) ? {} : obj;
    }
    /**
     * It takes a filter object, picks only the valid keys, and then recursively filters the object until
     * it's empty
     * @param {object} filterObj - object - The object that you want to filter.
     */
    filterRecursive(filterObj) {
        var _a, _b, _c, _d;
        const validFilterObj = _.pick(filterObj, [
            ...this.entity.filterableColumns,
            'AND',
            'NOT',
            'OR',
            ...this.entity.filterableReltionship,
        ]);
        for (const key in validFilterObj) {
            const filterValue = validFilterObj[key];
            if (typeof filterValue != 'string' && !Array.isArray(filterValue)) {
                const trimmedFilterValue = _.pick(filterValue, [
                    ...this.validStringFilter,
                    ...this.entity.filterableReltionshipColumns,
                ]);
                // console.log('0', { filterValue, trimmedFilterValue, key })
                validFilterObj[key] = (0, query_builder_constant_1.isObjectEmpty)(trimmedFilterValue)
                    ? undefined
                    : trimmedFilterValue;
                if ((_a = validFilterObj[key]) === null || _a === void 0 ? void 0 : _a.in)
                    validFilterObj[key].in = this.toArray((_b = validFilterObj[key]) === null || _b === void 0 ? void 0 : _b.in.split(','));
                if ((_c = validFilterObj[key]) === null || _c === void 0 ? void 0 : _c.notIn)
                    validFilterObj[key].notIn = this.toArray((_d = validFilterObj[key]) === null || _d === void 0 ? void 0 : _d.notIn.split(','));
            }
        }
        return validFilterObj;
    }
}
exports.default = ApiQueryBuilder;
//# sourceMappingURL=query-builder.js.map